library stencil;

String render(Component comp) => comp.render();

/// Base class for a Stencil component
///
/// Use [render] method to render the compoenent
///
/// Extend render method to implement templates
///
/// Use [forEach] in templates to
abstract class Component {
  /// Render lists based on [Iterable]
  String forEach<T>(Iterable<T> list, String renderer(T item)) =>
      list.map(renderer).join();

  /// Render lists based on ranges
  String range(int start, int end, String renderer(int index),
          [int step = 1]) =>
      _range(start, end, step).map(renderer).join();

  /// Render lists based on ranges from 0 to arbitrary positive integer
  String until(int end, String renderer(int index), [int step = 1]) =>
      _rangeUntil(end, step).map(renderer).join();

  /// Render lists based on iterating over a [Map]
  String map<KT, VT>(
      Map<KT, VT> map, String renderer(KT key, VT v, int index)) {
    final sb = new StringBuffer();

    int i = 0;
    for (KT key in map.keys) {
      sb.write(renderer(key, map[key], i));
      i++;
    }

    return sb.toString();
  }

  /// Conditional rendering
  String when(bool cond, String renderer(),
      [String elseRenderer() = nopRender]) {
    if (cond) return renderer();
    return elseRenderer();
  }

  /// Render another component
  String comp(Component c) => c.render();

  /// Implements the actual rendering
  String render();

  /// A Simple no-operation renderer
  static String nopRender() => '';

  /// Iterable geenrator for ranging
  static Iterable<int> _range(int start, int end, [int step = 1]) sync* {
    if (step <= 0)
      throw new ArgumentError.value(step, 'step', 'Must be greater than 1!');

    if (start < end) {
      for (int i = start; i < end; i += step) {
        yield i;
      }
    } else {
      for (int i = start; i > end; i -= step) {
        yield i;
      }
    }
  }

  /// Iterable generator for until-ranging
  static Iterable<int> _rangeUntil(int end, [int step = 1]) sync* {
    if (step <= 0)
      throw new ArgumentError.value(step, 'step', 'Must be greater than 1!');

    if (end < 0) throw new ArgumentError.value(end, 'end', 'Must be positive!');

    for (int i = 0; i < end; i += step) {
      yield i;
    }
  }
}
